import json
from itertools import chain

from scrapy.utils.spider import arg_to_iter

from scrapely.extraction.pageparsing import (
    TemplatePageParser, _AUTO_CLOSE_TAGS_ON_OPEN, _END_UNPAIREDTAG_TAGS,
    HtmlTag
)
from scrapely.extraction.pageobjects import (
    TemplatePage, AnnotationTag, AnnotationText
)


class SlybotTemplatePageParser(TemplatePageParser):
    def to_template(self, descriptors=None):
        if self.labelled_tag_stacks:
            tags = sorted(
                filter(bool, chain(*self.labelled_tag_stacks.values())),
                key=lambda a: a.start_index, reverse=True)
            try:
                next_tag_index = self.annotations[-1].end_index + 1
            except IndexError:
                next_tag_index = self.next_tag_index - len(tags) - 2
            for tag in tags:
                tag.end_index = next_tag_index
                next_tag_index += 1
            self.annotations.extend(tags)
        if descriptors is None:
            descriptors = {}
        return SlybotTemplatePage(self.html_page, self.token_dict,
                                  self.token_list, self.annotations,
                                  self.html_page.page_id, self.ignored_regions,
                                  self.extra_required_attrs, descriptors)

    @staticmethod
    def _read_template_annotation(html_tag):
        template_attr = html_tag.attributes.get('data-scrapy-annotate')
        if template_attr is None:
            return None
        try:
            return json.loads(template_attr.replace('&quot;', '"'))
        except ValueError:
            # JSON may be invalid in the case of extremely malformed html
            return None

    def read_jannotations(self, html_tag):
        jannotations = self._read_template_annotation(html_tag)
        return jannotations if jannotations else None

    def build_annotation(self, jannotation, is_open=True):
        annotation = AnnotationTag(self.next_tag_index,
                                   self.next_tag_index + 1)
        content_key = jannotation.pop('text-content', 'content')
        attribute_annotations = jannotation.pop('annotations', {})
        content = attribute_annotations.pop(content_key, None)
        if is_open and content:
            annotation.surrounds_attribute = content
        annotation.tag_attributes = list(attribute_annotations.items())
        annotation.metadata = jannotation
        return annotation

    def handle_generated(self, annotation, ignored=False):
        if not annotation.metadata.pop('generated', False):
            return
        self.token_list.pop()
        annotation.start_index -= 1
        if self.previous_element_class == HtmlTag:
            annotation.annotation_text = AnnotationText('')
        else:
            annotation.annotation_text = AnnotationText(self.prev_data)
        if ignored:
            ignored = self.ignored_regions.pop()
            self.ignored_regions.append((ignored[0] - 1, ignored[1]))

    def handle_variant(self, annotation, is_open=True):
        try:
            variant_id = int(annotation.metadata.pop('variant', 0))
        except (TypeError, ValueError):
            variant_id = None
        if variant_id is not None and variant_id > 0:
            if is_open and annotation.surrounds_attribute is not None:
                self.variant_stack.append(variant_id)
            else:
                annotation.variant_id = variant_id
        if is_open and self.variant_stack and annotation.variant_id is None:
            variant_id = self.variant_stack[-1]
            if variant_id == '0':
                variant_id = None
            annotation.variant_id = variant_id

    def handle_ignore(self, html_tag, is_open=True):
        next_tag_idx = self.next_tag_index
        read_bool = lambda x: self._read_bool_template_attribute(html_tag, x)
        ignore = read_bool('ignore')
        ignore_beneath = read_bool('ignore-beneath')
        if ignore:
            if html_tag.tag == "img":
                self.ignored_regions.append((next_tag_idx, next_tag_idx + 1))
            elif is_open:
                self.ignored_regions.append((next_tag_idx, None))
                self.ignored_tag_stacks[html_tag.tag].append(html_tag)
        elif is_open and self.ignored_tag_stacks.get(html_tag.tag):
            self.ignored_tag_stacks[html_tag.tag].append(None)
        if ignore_beneath:
            self.ignored_regions.append((next_tag_idx, None))
        return ignore or ignore_beneath

    def handle_replacement(self, html_tag):
        replacement = html_tag.attributes.pop("data-scrapy-replacement", None)
        if replacement:
            self.token_list.pop()
            self._add_token(replacement, html_tag.tag_type, html_tag.start,
                            html_tag.end)
            self.replacement_stacks[html_tag.tag].append(replacement)
        elif html_tag.tag in self.replacement_stacks:
            self.replacement_stacks[html_tag.tag].append(None)
        if self.unpairedtag_stack:
            if html_tag.tag in _END_UNPAIREDTAG_TAGS:
                self._close_unpaired_tag()
            else:
                self.unpairedtag_stack.append(html_tag.tag)
        tagname = replacement or self._update_replacement_stack(html_tag)
        self._handle_unclosed_tags(tagname, _AUTO_CLOSE_TAGS_ON_OPEN)
        return tagname

    def _handle_unpaired_tag(self, html_tag):
        self.handle_ignore(html_tag, is_open=False)
        jannotations = self.read_jannotations(html_tag)
        for jannotation in arg_to_iter(jannotations):
            if self.unpairedtag_stack:
                self._close_unpaired_tag()
            self.extra_required_attrs.extend(jannotation.pop('required', []))
            annotation = self.build_annotation(jannotation)
            self.handle_variant(annotation, is_open=False)
            self.annotations.append(annotation)
        self.next_tag_index += 1

    def _handle_open_tag(self, html_tag):
        ignored = self.handle_ignore(html_tag)
        tagname = self.handle_replacement(html_tag)
        jannotations = self.read_jannotations(html_tag)
        if not jannotations and tagname in self.labelled_tag_stacks:
            # add this tag to the stack to match correct end tag
            self.labelled_tag_stacks[tagname].append(None)
        increment = not jannotations
        for jannotation in arg_to_iter(jannotations):
            self.extra_required_attrs.extend(jannotation.pop('required', []))
            annotation = self.build_annotation(jannotation)
            self.handle_generated(annotation, ignored)
            self.handle_variant(annotation)
            # Don't increment generated/text annotation
            if annotation.annotation_text is None and not increment:
                increment = True
            # look for a closing tag if the content is important
            if annotation.surrounds_attribute:
                self.labelled_tag_stacks[tagname].append(annotation)
            else:
                annotation.end_index = annotation.start_index + 1
                self.annotations.append(annotation)
        self.next_tag_index += increment


class SlybotTemplatePage(TemplatePage):
    __slots__ = ('descriptors', 'modifiers')

    def __init__(self, htmlpage, token_dict, page_tokens, annotations,
                 template_id=None, ignored_regions=None, extra_required=None,
                 descriptors=None):
        self.descriptors = descriptors
        self.modifiers = {}
        for descriptor in descriptors.values():
            self.modifiers.update(getattr(descriptor, 'extractors', {}))
        super(SlybotTemplatePage, self).__init__(
            htmlpage, token_dict, page_tokens, annotations, template_id,
            ignored_regions, extra_required)

    def descriptor(self, descriptor_name=None):
        if descriptor_name is None:
            descriptor_name = '#default'
        return self.descriptors.get(descriptor_name, {})


def parse_template(token_dict, template_html, descriptors):
    """Create an TemplatePage object by parsing the annotated html."""
    parser = SlybotTemplatePageParser(token_dict)
    parser.feed(template_html)
    return parser.to_template(descriptors)



import copy

import six

from collections import defaultdict
from itertools import chain

from parsel import SelectorList
from scrapely.extraction.regionextract import TextRegionDataExtractor
from scrapely.htmlpage import HtmlPageParsedRegion, HtmlPageRegion
from scrapely.extractors import htmlregion
from scrapy.utils.spider import arg_to_iter
from slybot.item import SlybotFieldDescriptor

from w3lib.html import remove_tags

from .utils import (cached_property, load_annotations, region_id,
                    _DEFAULT_EXTRACTOR)
from .exceptions import ItemNotValidError, MissingRequiredError


def _compose(f, g):
    """given unary functions f and g, return a function that computes f(g(x))
    """
    def _exec(x):
        ret = g(x)
        if ret is not None:
            ret = HtmlPageRegion(ret.htmlpage, remove_tags(ret.text_content))
            return f(ret)
        return None
    return _exec


class ItemProcessor(object):
    """Processor for extracted data."""
    ignore = False
    should_overwrite = True
    selector_modes = (u'css', u'xpath')

    def __init__(self, data, extractor, regions, parent_region=None,
                 htmlpage=None):
        self.annotation = extractor.annotation
        self.id = self.annotation.metadata.get(u'id')
        self.regions = arg_to_iter(regions)
        parent_region = arg_to_iter(parent_region) if parent_region else []
        self.parent_region = parent_region
        self.modifiers = extractor.modifiers or {}
        self.schema = extractor.schema or {}
        if hasattr(htmlpage, u'htmlpage'):
            htmlpage = htmlpage.htmlpage
        self.htmlpage = htmlpage
        self.annotations = list(
            load_annotations(getattr(extractor, 'extractors', [])))
        self.fields = self._process_fields(data)

    @cached_property
    def field(self):
        """Field display name."""
        return getattr(self.descriptor, 'description', self.name)

    @cached_property
    def descriptor(self):
        """Field descriptor and adaptor."""
        return getattr(self.schema, u'attribute_map', {}).get(self.name)

    @cached_property
    def name(self):
        """Field unique name."""
        return self.annotation.metadata.get(u'field')

    @cached_property
    def description(self):
        """Field display name."""
        return self.field

    @property
    def region_id(self):
        return ', '.join(str(r) for r in self.regions)

    @cached_property
    def metadata(self):
        return self.annotation.metadata

    def attribute_query(self, metadata):
        """Extract attribute or content from a region."""
        content_field = metadata.get(u'text-content', u'content')
        attribute = metadata.get(u'attribute', content_field)
        if attribute == content_field:
            return u'self::node()'
        return u'@%s' % attribute

    def _process_fields(self, data):
        """Convert extracted data into ItemField fields."""
        schema, modifiers, page = self.schema, self.modifiers, self.htmlpage
        fields = {}
        for field_num, (field, value) in enumerate(self._normalize_data(data)):
            # Repeated Fields and nested items support
            if hasattr(field, 'fields'):
                self._add_item(field, fields)
                continue
            # New style annotation field
            elif isinstance(field, dict):
                field_id = field.get(u'id') or field_num
            # Legacy attribute, field mapping annotation
            else:
                field_id = field_num
                field = {u'field': field, u'id': field_id,
                         u'attribute': u'content'}
            field_cls = self._field_class(field['field'])
            fields[field_id] = field_cls(value, field, schema, modifiers, page)
        return fields

    def _field_class(self, field):
        if not field or field.startswith('#'):
            return ScrapelyField
        if field.startswith('_'):
            return ProcessedField
        return ItemField

    def _add_item(self, item, fields=None):
        fields = self.fields if fields is None else fields
        child = None
        if len(item.fields) == 1:
            child = next(six.itervalues(item.fields))
        if child and item.descriptor == child.extractor:
            fields[child.id] = child
        else:
            fields[item.name] = item

    def _normalize_data(self, data):
        """Normalize extracted data for conversion into ItemFields."""
        if isinstance(data, dict):
            data = data.items()
        elif data and not isinstance(data[0], (tuple, dict)):
            data = [data]
        for i in data:
            if hasattr(i, u'items'):
                i = i.items()
            else:
                i = (i,)
            other_fields = []
            for fields in chain(arg_to_iter(i), other_fields):
                try:
                    fields, value = fields
                except ValueError:
                    for field in fields:
                        if hasattr(field, 'fields'):
                            yield field, None
                        elif len(field) == 2:
                            # Queue repeated fields for normalization
                            other_fields.append(field)
                    continue
                if isinstance(fields, list):
                    # More than a one attribute for a single annotation
                    for field in fields:
                        yield field, value
                elif isinstance(fields, six.string_types):
                    # Legacy field support
                    yield {u'field': fields, u'attribute': u'content'}, value
                else:
                    yield fields, value

    def process(self, selector=None, include_meta=False):
        """Extract CSS and XPath annotations and dump item."""
        if selector is not None:
            self._process_selectors(selector)
        return self.dump(include_meta, validate=True)

    def _process_selectors(self, selector):
        selector_annotations = self._selector_annotations()
        field_annotations = (
            f.metadata for f in self.fields.values()
            if hasattr(f, 'selection_mode') and
            f.selection_mode in self.selector_modes
        )
        for field_id, field in self.fields.items():
            if hasattr(field, 'fields'):
                value = field.process(selector)
                if not value:
                    continue
                meta = {'id': field.id, 'field': field.name,
                        'attribute': 'content'}
                self.fields[field_id] = ProcessedField(
                    value, meta, self.schema,
                    self.modifiers, self.htmlpage)
        all_selector_annotations = list(
            chain(selector_annotations, field_annotations))
        if all_selector_annotations:
            self._process_css_and_xpath(all_selector_annotations, selector)

    def _selector_annotations(self):
        for annotation in self.annotations:
            meta = annotation.metadata
            if meta.get(u'selection_mode') not in self.selector_modes:
                continue
            surrounds = arg_to_iter(annotation.surrounds_attribute) or []
            tags = chain(*(a for _, a in annotation.tag_attributes))
            for attribute in chain(surrounds, tags):
                new_attribute = {k: v for k, v in meta.items()}
                new_attribute.update(attribute)
                yield new_attribute

    def _process_css_and_xpath(self, annotations, selector):
        schema, modifiers, page = self.schema, self.modifiers, self.htmlpage
        region_ids = list(filter(bool, (region_id(r) for r in self.regions)))
        query = ','.join(('[data-tagid="%s"]' % rid for rid in region_ids))
        parents = {e._root for e in selector.css(query)}
        containers = ()
        if self.parent_region:
            pquery = '[data-tagid="%s"]' % self.parent_region
            containers = {e._root for e in selector.css(pquery)}
        for i, a in enumerate(annotations, start=len(self.fields)):
            mode = a.get(u'selection_mode')
            query = a.get(mode if mode != 'css' else u'selector')
            try:
                elems = self._pick_elems(
                    getattr(selector, mode)(query), parents, containers)
            except ValueError:
                continue
            for elem in elems:
                elem._root.attrib.pop('data-tagid', None)
            extracted = elems.xpath(self.attribute_query(a)).extract()
            value = list(map(six.text_type.strip, extracted))
            if value:
                aid = a.get(u'id') or i
                value = [htmlregion(v) for v in arg_to_iter(value)]
                self.fields[aid] = ItemField(value, a, schema, modifiers, page)

    def _pick_elems(self, elements, parents, containers):
        closest_elements, closest_set = SelectorList(), set()
        other_elements = SelectorList()
        for element in elements:
            try:
                element_parents = element._root.iterancestors()
            except AttributeError:
                continue
            for parent in element_parents:
                if parent in parents:
                    closest_elements.append(element)
                    closest_set.add(element)
                if parent in containers and element not in closest_set:
                    break
            else:
                other_elements.append(element)
        if closest_elements:
            return closest_elements
        elif other_elements:
            return other_elements
        return elements

    def merge(self, other):
        """Merge this instance with another ItemProcessor instance

        Add additional regions.
        Add additional annotations.
        Add new fields from the other ItemProcessor.
        Merge existing field values.
        """
        for region in other.regions:
            if region not in self.regions:
                self.regions.append(region)
        aids = {a.metadata.get(u'id') for a in self.annotations}
        other_aids = {a.metadata.get(u'id') for a in other.annotations}
        missing_ids = other_aids - aids
        for annotation in other.annotations:
            id_ = annotation.metadata.get(u'id')
            if id_ and id_ in missing_ids:
                self.annotations.append(annotation)
        if other.name:
            self._add_item(other)
            return
        for field_id, field in other.fields.items():
            if field_id in self.fields:
                self.fields[field_id].merge(field)
            else:
                self.fields[field_id] = field

    def dump(self, include_meta=False, validate=False):
        """Dump processed fields into a new item."""
        try:
            return self._dump(include_meta, validate)
        except (MissingRequiredError, ItemNotValidError):
            return {}

    def _dump(self, include_meta=False, validate=False):
        item = defaultdict(list)
        meta = defaultdict(dict)
        schema_id = getattr(self.schema, u'id', None)
        for field in self.fields.values():
            value = field.dump()
            if not value or field.ignore:
                continue
            if field.should_overwrite:
                item[field.field] = value
            else:
                item[field.field].extend(value)
            if include_meta:
                meta[field.id].update(dict(schema=schema_id, **field.metadata))
        if validate:
            self._validate(item)

        # Rename fields to their human readable names
        item = self._item_with_names(item)
        if include_meta:
            item[u'_meta'] = meta
        if u'_type' not in item:
            _type = getattr(self.schema, u'description', schema_id)
            if _type:
                item[u'_type'] = _type
        return item

    def _validate(self, item):
        item_fields = self._item_with_names(item, u'name')
        # Bring keys from nested items into primary item for required check
        for key, value in list(item_fields.items()):
            if isinstance(value, dict):
                for sub_key in value:
                    item_fields[sub_key] = True
        # Check if a pre prcessed item has been provided
        if u'_type' in item_fields:
            return item_fields
        if (hasattr(self.schema, u'_item_validates') and
                not self.schema._item_validates(item_fields)):
            raise ItemNotValidError
        if all(fname[0] == '_' or fname == 'url' for fname in item_fields):
            raise ItemNotValidError

    def _item_with_names(self, item, attribute=u'description'):
        item_dict = {}
        for field, value in item.items():
            if not (field and value):
                continue
            if hasattr(field, attribute):
                key = getattr(field, attribute)
                if getattr(field, 'should_overwrite', False):
                    item_dict[key] = value
                else:
                    item_dict[key] = [
                        v for v in chain(arg_to_iter(item_dict.get(key, [])),
                                         arg_to_iter(value))
                    ]
            else:
                item_dict[field] = value
        return item_dict

    def __getitem__(self, key):
        values = []
        for field in self.fields.values():
            if hasattr(field.field, u'get'):
                field_name = field.field.get(u'field')
            else:
                field_name = field.field
            if field_name == key:
                values.extend(field.dump())
        return values

    def __bool__(self):
        return bool(self.dump())

    __nonzero__ = __bool__

    def __len__(self):
        return len(self.fields)

    def __hash__(self):
        return hash(str(self.id) + str(self.region_id))

    def __setitem__(self, key, value):
        self.fields[key] = ProcessedField(value, {u'id': key, u'field': key,
                                                  u'attribute': u'content'})

    def __str__(self):
        return u'%s, %s' % (self.id, self.region_id)

    def __repr__(self):
        return u'%s(%s, %s)' % (self.__class__.__name__, str(self),
                                repr(self.fields))


class ItemField(object):
    should_overwrite = False

    def __init__(self, value, meta, schema=None, modifiers=None,
                 htmlpage=None):
        self.htmlpage = htmlpage
        self.value = value
        self._meta = meta
        self.id = meta.get(u'id')
        self._field = meta[u'field']
        self.attribute = meta[u'attribute']
        self.selection_mode = meta.get(u'selection_mode', u'auto')
        self.extractor, self.adaptors = self._load_extractors(
            self._field, schema, modifiers)

    @cached_property
    def field(self):
        return self

    @cached_property
    def ignore(self):
        if not self._field:
            return True
        return False

    @cached_property
    def description(self):
        """Field display name."""
        return getattr(self.extractor, u'description', self._field)

    @cached_property
    def name(self):
        """Field unique name."""
        return getattr(self.extractor, u'name', self._field)

    @cached_property
    def metadata(self):
        meta = {k: v for k, v in self._meta.items() if k not in
                (u'name', u'value', u'schema')}
        meta['field'], meta['value'], meta['processed'] = (
            self._field, self.value, self.dump())
        return meta

    @cached_property
    def required(self):
        return self._meta.get('required') or False

    def dump(self):
        """Process and adapt extracted data for field."""
        values = self._process()
        return self._adapt(values)

    def merge(self, other):
        try:
            self.value.extend(other.value)
        except AttributeError:
            self.value = other.value

    def _load_extractors(self, field, schema, modifiers):
        field, _meta = self._field, self._meta
        extractors = []
        try:
            field_extraction = schema.attribute_map.get(field)
        except AttributeError:
            field_extraction = None
        if field_extraction is None:
            field_extraction = SlybotFieldDescriptor(field, field,
                                                     _DEFAULT_EXTRACTOR)
        if u'pre_text' in _meta or u'post_text' in _meta:
            text_extractor = TextRegionDataExtractor(
                _meta.get(u'pre_text', u''),
                _meta.get(u'post_text', u''))
            field_extraction = copy.deepcopy(field_extraction)
            field_extraction.extractor = _compose(
                field_extraction.extractor, text_extractor.extract)
        extractors = _meta.get(u'extractors', [])
        if isinstance(extractors, dict):
            extractors = extractors.get(field, [])
        adaptors = []
        for extractor in extractors:
            if extractor in modifiers:
                adaptors.append(modifiers[extractor])
        return field_extraction, adaptors

    def _process(self):
        values = []
        for value in arg_to_iter(self.value):
            if (isinstance(value, (HtmlPageParsedRegion, HtmlPageRegion)) and
                    hasattr(self.extractor, u'extractor')):
                value = self.extractor.extractor(value)
            if value:
                values.append(value)
        if hasattr(self.extractor, u'adapt'):
            values = [self.extractor.adapt(x, self.htmlpage) for x in values
                      if x and not isinstance(x, (dict, ItemProcessor))]
        else:
            values = list(filter(bool, values))
        return values

    def _adapt(self, values):
        for adaptor in self.adaptors:
            if values:
                values = [adaptor(v, self.htmlpage) for v in values if v]
        if self.required and not values:
            raise MissingRequiredError
        return values

    def __hash__(self):
        return hash(str(self.id) + str(self._field))

    def __str__(self):
        return u'%s: %s | id=%s' % (self.description, self.dump(), self.id)

    def __repr__(self):
        return u'%s(%s, field=%s, extractor=%s, adaptors=%s)' % (
            self.__class__.__name__, str(self), self._field, self.extractor,
            self.adaptors)


class ProcessedField(ItemField):
    """Field without any transformations."""
    should_overwrite = True

    @cached_property
    def field(self):
        return self._field

    def dump(self):
        return self.value


class ScrapelyField(ItemField):
    """Field used for guiding extraction but not used for extracting data."""
    ignore = True



"""
Link extraction for auto scraping
"""
import six
from scrapy.link import Link
from scrapy.selector import Selector

from slybot.linkextractor.base import BaseLinkExtractor
RSS_XPATH = "//item/link/text()"
SITEMAP_XPATH = "//urlset/url/loc/text()|//sitemapindex/sitemap/loc/text()"
ATOM_XPATH = "//link/@href"


class XmlLinkExtractor(BaseLinkExtractor):
    """Link extractor for XML sources"""
    def __init__(self, xpath=None, **kwargs):
        if not xpath:
            xpath = '|'.join((RSS_XPATH, SITEMAP_XPATH, ATOM_XPATH))
            self.remove_namespaces = True
        else:
            self.remove_namespaces = kwargs.pop('remove_namespaces', False)
        super(XmlLinkExtractor, self).__init__(**kwargs)
        self.xpath = xpath

    def _extract_links(self, response):
        body = response.body_as_unicode()
        _type = 'html'
        if body.lstrip().startswith('<?xml version='):
            _type = 'xml'
        xxs = Selector(text=body, type=_type)
        if self.remove_namespaces:
            xxs.remove_namespaces()
        for url in xxs.xpath(self.xpath).extract():
            if not isinstance(url, six.text_type):
                url = url.encode(response.encoding)
            yield Link(url)


class RssLinkExtractor(XmlLinkExtractor):
    """Link extraction from RSS feeds"""
    def __init__(self, **kwargs):
        super(RssLinkExtractor, self).__init__(RSS_XPATH, **kwargs)


class SitemapLinkExtractor(XmlLinkExtractor):
    """Link extraction for sitemap.xml feeds"""
    def __init__(self, **kwargs):
        kwargs['remove_namespaces'] = True
        super(SitemapLinkExtractor, self).__init__(SITEMAP_XPATH, **kwargs)


class AtomLinkExtractor(XmlLinkExtractor):
    def __init__(self, **kwargs):
        kwargs['remove_namespaces'] = True
        super(AtomLinkExtractor, self).__init__(ATOM_XPATH, **kwargs)



import json
import six

from scrapy import Selector
from scrapy.utils.spider import arg_to_iter
from scrapely.htmlpage import parse_html, HtmlTag, HtmlDataFragment

from collections import defaultdict
from itertools import tee, count, groupby, chain
from operator import itemgetter

from slybot.utils import (serialize_tag, add_tagids, remove_tagids, TAGID,
                          OPEN_TAG, CLOSE_TAG, UNPAIRED_TAG, GENERATEDTAGID)

from .migration import _get_parent, short_guid


class Annotations(object):
    def __init__(self, sample, **options):
        self.sample = sample
        plugins = sample.setdefault('plugins', {})
        self.data = plugins.setdefault('annotations-plugin', {'extracts': []})
        self.annotations = self.data['extracts']
        self.legacy = options.get('legacy') or False
        self.data['extracts'] = _clean_annotation_data(self.annotations)
        self._elems = {}

    @property
    def html(self):
        if hasattr(self, '_html'):
            return self._html
        template = self.sample
        body = template.get('body') or 'original_body'
        if body not in template:
            if 'original_body' in template:
                body = 'original_body'
            else:
                bodies = [k for k, v in template.items()
                          if v and k.endswith('_body')]
                if bodies:
                    body = bodies[0]
        self._html = template[body]
        return self._html

    @property
    def selector(self):
        if hasattr(self, '_selector'):
            return self._selector
        self._selector = Selector(text=self.numbered_html)
        return self._selector

    @property
    def numbered_html(self):
        if hasattr(self, '_numbered_html'):
            return self._numbered_html
        self._numbered_html = add_tagids(self.html)
        return self._numbered_html

    def build(self):
        if not self.annotations and 'annotated_body' in self.sample:
            return self.data
        self.sample['annotated_body'] = self.apply()
        return self.data

    def verify(self, tagid_annotations):
        def on_cid(a):
            return a['id'] if a.get('item_container') else a['container_id']
        containers = {a['id']: a for a in chain(*tagid_annotations)
                      if a.get('item_container')}
        missing = {}
        groups = groupby(sorted(chain(*tagid_annotations), key=on_cid), on_cid)
        for group, annotations in groups:
            if group not in containers:
                annotations = list(annotations)
                missing[group] = list(annotations)
        if missing:
            all_containers = {a['id']: a for a in self.annotations
                              if a.get('item_container')}
            for container_id, annnotations in missing.items():
                container = all_containers[container_id]
                annotations = list(annotations)
                elements = list(chain(*(self.elements(a)
                                        for a in annotations)))
                parent = _get_parent(elements, self.selector)
                container['tagid'] = parent.attrib.get(TAGID)
                tagid_annotations.append(container)
        return tagid_annotations

    def generate(self, annotations):
        data = {}
        annotation_data = []
        for annotation in arg_to_iter(annotations):
            if 'annotations' in annotation:
                annotation_data.append({
                    'id': annotation.get('id', short_guid()),
                    'annotations': annotation.get('annotations', {}),
                    'required': annotation.get('required', []),
                    'required_fields': annotation.get('required', []),
                    'variant': int(annotation.get('variant', 0)),
                    'generated': annotation.get('generated', False),
                    'text-content': annotation.get('text-content', 'content'),
                    'item_container': annotation.get('item_container', False),
                    'container_id': annotation.get('container_id'),
                    'schema_id': annotation.get('schema_id'),
                    'repeated': annotation.get('repeated'),
                    'siblings': annotation.get('siblings'),
                    'field': annotation.get('field'),
                    'selector': annotation.get('selector'),
                    'selection_mode': annotation.get('selection_mode'),
                    'min_jump': annotation.get('min_jump', -1),
                    'max_separator': annotation.get('max_separator', -1),
                    'xpath': annotation.get('xpath')
                })
            if 'ignore' in annotation or 'ignore_beneath' in annotation:
                if annotation.get('ignore_beneath'):
                    data['data-scrapy-ignore-beneath'] = 'true'
                elif annotation.get('ignore'):
                    data['data-scrapy-ignore'] = 'true'
        if annotation_data:
            if self.legacy:
                annotation_data = annotation_data[0]
            serialized = json.dumps(annotation_data).replace('"', '&quot;')
            data['data-scrapy-annotate'] = serialized
        return data

    def _generate_elem(self, annotation, text):
        sections = ['<ins']
        annotation_info = self.generate(annotation)
        annotation_info[GENERATEDTAGID] = annotation.get('id')
        attributes = []
        for key, value in annotation_info.items():
            attributes.append('%s="%s"' % (key, value))
        sections.append(' '.join(attributes))
        if len(sections) > 1:
            sections[0] += ' '
        sections.extend(['>', text, '</ins>'])
        return ''.join(sections)

    def _get_generated(self, element, annotations, nodes, inserts):
        html_body = self.numbered_html
        eid = insert_after_tag = _get_data_id(element)
        text_strings = _get_text_nodes(nodes, html_body)
        text_content = ''.join((s.lstrip() for s in text_strings))
        pre_selected = []
        for annotation in annotations:
            start, end = _get_generated_slice(annotation)
            pre_selected.append(
                (text_content[0:start], text_content[start:end], annotation))
        stack = [insert_after_tag]
        next_text_node = ''
        for i, node in enumerate(nodes):
            if isinstance(node, HtmlTag):
                if node.tag_type == OPEN_TAG:
                    tagid = node.attributes.get(TAGID, '').strip()
                    if tagid:
                        stack.append(tagid)
                elif node.tag_type == CLOSE_TAG and stack:
                    insert_after_tag = stack.pop()
            elif (isinstance(node, HtmlDataFragment) and len(stack) == 1):
                text = html_body[node.start:node.end]
                # This allows for a clean way to insert fragments up until the
                # next tag in apply_annotations if we have already inserted a
                # new generated tag
                if not node.is_text_content and inserts.get(insert_after_tag):
                    inserts[insert_after_tag].append(text)
                    continue
                removed = 0
                inserted = False
                for j, (pre, selected, anno) in enumerate(pre_selected[:]):
                    if selected and selected in text:
                        previous, post = text.split(selected, 1)
                        if previous.strip() in pre:
                            pre_selected.pop(j - removed)
                            removed += 1
                            generated = self._generate_elem(anno, selected)
                            # Next immediate text node will be returned and
                            # added to the new document. Other text nodes
                            # within this node will be added after other child
                            # nodes have been closed.
                            if (insert_after_tag == eid and
                                    not anno.get('insert_after')):
                                next_text_node += previous + generated
                                inserted = True
                            else:
                                inserts[insert_after_tag].extend([previous,
                                                                  generated])
                            text = post
                if inserted:
                    next_text_node += text
                else:
                    inserts[insert_after_tag].append(text)
        return next_text_node

    def apply(self):
        selector_annotations, tagid_annotations = self.split()
        inserts, numbered_html = defaultdict(list), self.numbered_html
        if selector_annotations:
            converted_annotations = self.apply_selector(selector_annotations)
            tagid_annotations += converted_annotations
        if not self.legacy:
            tagid_annotations = self.verify(tagid_annotations)
        target = iter(parse_html(numbered_html))
        output, stack = [], []
        elem = next(target)
        last_id = 0
        # XXX: A dummy element is added to the end so if the last annotation is
        #      generated it will be added to the output
        filtered = defaultdict(list)
        for grouped in tagid_annotations:
            for ann in arg_to_iter(grouped):
                filtered[ann['tagid']].append(ann)
        dummy = [(1e9, [{}])]
        sorted_annotations = sorted([(int(k), v) for k, v in filtered.items()
                                     if k is not None])
        try:
            for aid, annotation_data in chain(sorted_annotations, dummy):
                # Move target until replacement/insertion point
                while True:
                    while not isinstance(elem, HtmlTag) or elem.tag == 'ins':
                        output.append(numbered_html[elem.start:elem.end])
                        elem = next(target)
                    if elem.tag_type in {OPEN_TAG, UNPAIRED_TAG}:
                        last_id = elem.attributes.get(TAGID)
                        stack.append(last_id)
                    if elem.tag_type in {CLOSE_TAG, UNPAIRED_TAG} and stack:
                        if ('__added' not in elem.attributes and
                                last_id is not None and aid is not None and
                                int(last_id) < int(aid)):
                            output.append(numbered_html[elem.start:elem.end])
                            elem.attributes['__added'] = True
                        last_inserted = stack.pop()
                        to_insert = inserts.pop(last_inserted, None)
                        if to_insert:
                            output.extend(to_insert)
                            # Skip all nodes up to the next HtmlTag as these
                            # have already been added
                            while True:
                                elem = next(target)
                                try:
                                    last_id = elem.attributes.get(TAGID,
                                                                  last_id)
                                except AttributeError:
                                    pass
                                if isinstance(elem, HtmlTag):
                                    break
                            continue
                    if (last_id is not None and aid is not None and
                            int(last_id) < int(aid)):
                        if '__added' not in elem.attributes:
                            output.append(numbered_html[elem.start:elem.end])
                            elem.attributes['__added'] = True
                        elem = next(target)
                    else:
                        break

                generated = []
                next_generated = []
                regular_annotations = []
                # Place generated annotations at the end and sort by slice
                for annotation in sorted(annotation_data, key=_annotation_key):
                    if annotation.get('generated'):
                        if annotation.get('insert_after'):
                            next_generated.append(annotation)
                        else:
                            generated.append(annotation)
                    else:
                        regular_annotations.append(annotation)
                # Add annotations data as required
                if regular_annotations:
                    annotation_info = self.generate(regular_annotations)
                    for key, val in annotation_info.items():
                        elem.attributes[key] = val
                next_text_section = ''
                if generated:
                    inner_data, target = tee(target)
                    nodes = _get_inner_nodes(inner_data)
                    next_text_section = self._get_generated(
                        elem, generated, nodes, inserts)
                if next_generated:
                    inner_data, target = tee(target)
                    open_tags = 0 if elem.tag_type == UNPAIRED_TAG else 1
                    nodes = _get_inner_nodes(inner_data, open_tags=open_tags,
                                             insert_after=True)
                    next_text_section = self._get_generated(
                        elem, next_generated, nodes, inserts)

                if '__added' not in elem.attributes:
                    output.append(serialize_tag(elem))
                    elem.attributes['__added'] = True
                # If an <ins> tag has been inserted we need to move forward
                if next_text_section:
                    while True:
                        elem = next(target)
                        if (isinstance(elem, HtmlDataFragment) and
                                elem.is_text_content):
                            break
                        output.append(numbered_html[elem.start:elem.end])
                    output.append(next_text_section)
        # Reached the end of the document
        except StopIteration:
            output.append(numbered_html[elem.start:elem.end])
        else:
            for element in target:
                output.append(numbered_html[element.start:element.end])
        return remove_tagids(''.join(output))

    def split(self):
        selector, tagid = [], []
        for ann in self.annotations:
            if ann:
                if ann.get('selector'):
                    selector.append(ann)
                elif ann.get('tagid') and (ann.get('annotations') or
                                           ann.get('ignore')):
                    tagid.append(ann)
        return selector, tagid

    def apply_selector(self, annotations):
        page = self.selector
        converted_annotations = []
        tagid_selector_map = {}
        added_repeated = {}
        containers = {}
        for annotation in annotations:
            if annotation.get('item_container'):
                containers[annotation['id']] = annotation
            selector = annotation.get('selector')
            elems = self.elements(annotation)
            if elems:
                tagid = min([int(e.attrib.get('data-tagid', 1e9))
                             for e in elems])
                annotation['tagid'] = tagid
                if selector:
                    tagid_selector_map[tagid] = selector
                converted_annotations.append(annotation)

            # Create container for repeated field annotation
            if (annotation.get('repeated') and
                    annotation.get('selection_mode') == 'auto' and
                    not annotation.get('item_container') and
                    elems is not None and len(elems) and
                    len(annotation.get('annotations')) == 1):
                repeated_parent = add_repeated_field(annotation, elems, page)
                if repeated_parent:
                    converted_annotations.append(repeated_parent)
                    container_id = repeated_parent['container_id']
                    added_repeated[container_id] = repeated_parent
        if added_repeated:
            for container_id, child in added_repeated.items():
                container = containers[container_id]
                if container.get('tagid') != child['tagid']:
                    continue
                elems = self.elements(container)
                parent = elems[0].getparent()
                container['tagid'] = int(parent.attrib.get('data-tagid', 1e9))
        return _merge_annotations_by_selector(converted_annotations)

    def elements(self, annotation):
        aid = annotation['id']
        if aid in self._elems:
            return self._elems[aid]
        selector = annotation.get('selector')
        if not selector:
            return []
        elems = []
        while selector and not elems:
            elems = [elem._root for elem in self.selector.css(selector)]
            selector = ' > '.join(selector.split(' > ')[1:])
        self._elems[aid] = elems or []
        if not elems:
            return []
        return elems


def _clean_annotation_data(data):
    result = []
    sticky_count, stickies = count(1), set()
    for ann in data:
        if ann.get('item_container'):
            ann['annotations'] = {'#portia-content': '#dummy'}
            ann['text-content'] = '#portia-content'
        elif 'data' in ann:
            modified_annotations = {}
            grp = itemgetter('attribute')
            for _id, value in ann['data'].items():
                value['id'] = '%s|%s' % (ann['id'], _id)
            sorted_annotations = sorted(ann['data'].values(), key=grp)
            for attribute, annotations in groupby(sorted_annotations, grp):
                modified_annotations[attribute] = list(annotations)
            ann['annotations'] = modified_annotations
        elif 'annotations' in ann:
            filtered_annotations = {}
            for k, v in ann['annotations'].items():
                if not v:
                    continue
                if v == '#sticky':
                    next_sticky = '_sticky%s' % next(sticky_count)
                    stickies.add(next_sticky)
                    v = next_sticky
                filtered_annotations[k] = v

            ann['annotations'] = filtered_annotations
            ann['required'] = list((set(ann.get('required', [])) | stickies) &
                                   set(filtered_annotations.values()))
        elif "ignore" in ann or "ignore_beneath" in ann:
            pass
        else:
            continue
        result.append(ann)
    return result


def _get_data_id(annotation):
    """Get id (a str) of an annotation."""
    if isinstance(annotation, HtmlTag):
        return annotation.attributes[TAGID]


def _get_text_nodes(nodes, html_body):
    text = []
    open_tags = 0
    for node in nodes:
        if isinstance(node, HtmlTag):
            if node.tag_type == OPEN_TAG:
                open_tags += 1
            elif node.tag_type == CLOSE_TAG:
                open_tags -= 1
        elif (isinstance(node, HtmlDataFragment) and
              node.is_text_content and open_tags == 0):
            text.append(html_body[node.start:node.end])
    return text


def _get_generated_slice(annotation):
    annotation_slice = annotation.get('slice', [0])[:2]
    if not annotation_slice:
        annotation_slice = [0, 0]
    elif len(annotation_slice) < 2:
        annotation_slice.append(annotation_slice[0])
    return annotation_slice


def _get_inner_nodes(target, open_tags=1, insert_after=False,
                     stop_on_next=False):
    nodes = []
    while open_tags > -0:
        elem = next(target)
        if isinstance(elem, HtmlTag):
            if elem.tag_type == OPEN_TAG:
                open_tags += 1
                if stop_on_next and elem.attributes.get(TAGID) is not None:
                    return nodes
            elif (stop_on_next and
                  elem.tag_type == UNPAIRED_TAG and
                  elem.attributes.get(TAGID) is not None):
                return nodes
            elif elem.tag_type == CLOSE_TAG:
                open_tags -= 1
        nodes.append(elem)
    if insert_after:
        return _get_inner_nodes(target, stop_on_next=True)
    return nodes


def _add_element(element, output, html):
    if '__added' not in element.attributes:
        output.append(html[element.start:element.end])
        element.attributes['__added'] = True
    return element


def _annotation_key(a):
    return a.get('generated', False) + sum(a.get('slice', []))


def _merge_annotations_by_selector(annotations):
    def grouper(x):
        return x.get('selector', '')
    annotations.sort(key=grouper)
    return [list(annos) for _, annos in groupby(annotations, key=grouper)]


def add_repeated_field(annotation, elems, page):
    parent = _get_parent(elems, page)
    field = next(six.itervalues(annotation['annotations']))[0]['field']
    container_id = '%s#parent' % annotation['id']
    if len(parent):
        tagid = int(parent.attrib.get('data-tagid', 1e9))
        parent_annotation = {
            'item_container': True,
            'id': container_id,
            'annotations': {'#portia-content': '#dummy'},
            'text-content': '#portia-content',
            'container_id': annotation['container_id'],
            'field': field,
            'tagid': tagid
        }
        annotation['item_container'] = True
        annotation['field'] = field
        annotation['container_id'] = container_id
        return parent_annotation



from six.moves.urllib_parse import urlparse
import chardet
import itertools
import json
import os
import re
import six

from collections import OrderedDict, namedtuple
from itertools import chain

from scrapely.htmlpage import HtmlPage, HtmlTagType
from scrapy.utils.misc import load_object
from w3lib.encoding import html_body_declared_encoding


TAGID = u"data-tagid"
GENERATEDTAGID = u"data-genid"
OPEN_TAG = HtmlTagType.OPEN_TAG
CLOSE_TAG = HtmlTagType.CLOSE_TAG
UNPAIRED_TAG = HtmlTagType.UNPAIRED_TAG
# Encodings: https://w3techs.com/technologies/overview/character_encoding/all
ENCODINGS = ['UTF-8', 'ISO-8859-1', 'Windows-1251', 'Shift JIS',
             'Windows-1252', 'GB2312', 'EUC-KR', 'EUC-JP', 'GBK', 'ISO-8859-2',
             'Windows-1250', 'ISO-8859-15', 'Windows-1256', 'ISO-8859-9',
             'Big5', 'Windows-1254', 'Windows-874']
MimeType = namedtuple('MimeType', ['type', 'maintype', 'subtype', 'params'])


def content_type(response):
    full_content_type = decode(response.headers.get('Content-Type') or u'')
    type_ = full_content_type.split(';', 1)
    split = type_[0].split('/', 1)
    if len(split) < 2:
        maintype = type_
        subtype = ''
    else:
        maintype, subtype = split
    # Parse params if needed
    return MimeType(full_content_type, maintype, subtype, [])


def encode(html, default=None):
    if isinstance(html, six.binary_type):
        return html
    return _encode_or_decode_string(html, type(html).encode, default)


def decode(html, default=None):
    if isinstance(html, six.text_type):
        return html
    return _encode_or_decode_string(html, type(html).decode, default)


def _encode_or_decode_string(html, method, default):
    if not default:
        encoding = html_body_declared_encoding(html)
        if encoding:
            default = [encoding]
        else:
            default = []
    elif isinstance(default, six.string_types):
        default = [default]
    for encoding in itertools.chain(default, ENCODINGS):
        try:
            return method(html, encoding)
        except (UnicodeDecodeError, UnicodeEncodeError, LookupError):
            pass
        except AttributeError:
            return html
    encoding = chardet.detect(html).get('encoding')
    return method(html, encoding)


def iter_unique_scheme_hostname(urls):
    """Return an iterator of tuples (scheme, hostname) over the given urls,
    filtering dupes
    """
    scheme_hostname = set()
    for x in urls:
        p = urlparse(x)
        scheme_hostname.add((p.scheme, p.hostname))
    return list(scheme_hostname)


def open_project_from_dir(project_dir):
    storage = Storage(project_dir)
    specs = {"spiders": SpiderLoader(storage)}
    for name in ['project', 'items', 'extractors']:
        try:
            specs[name] = storage.open('{}.json'.format(name))
        except IOError:
            specs[name] = {}
    return specs


def read(fp, encoding='utf-8'):
    content = fp.read()
    if hasattr(content, 'decode'):
        content = content.decode('utf-8')
    return content


def _build_sample(sample, legacy=False):
    from slybot.plugins.scrapely_annotations.builder import Annotations
    Annotations(sample, legacy=legacy).build()
    sample['page_id'] = sample.get('page_id') or sample.get('id') or ""
    sample['annotated'] = True
    return sample


def htmlpage_from_response(response, _add_tagids=False):
    body = response.body_as_unicode()
    if _add_tagids:
        body = add_tagids(body)
    return HtmlPage(response.url, response.headers, body,
                    encoding=response.encoding)


def load_plugins(settings):
    if settings.get('LOADED_PLUGINS', None):
        return settings.get('LOADED_PLUGINS', None)
    plugins = settings['PLUGINS']
    if plugins:
        return [load_object(p) if isinstance(p, str) else p for p in plugins]
    else:
        from slybot.plugins.scrapely_annotations import Annotations
        return [Annotations]


def load_plugin_names(settings):
    """
    Generate a unique name for a plugin based on the class name module name
    and path

    >>> settings = {'PLUGINS': ['a', 'b.c', 'a.c']}
    >>> load_plugin_names(settings)
    ['a', 'c', 'a.c']
    """
    seen = set()

    def generate_name(path, maxsplit=0, splits=None):
        if splits is None:
            splits = len(path.split('.')) - 1
        name = '.'.join(path.split('.', splits - maxsplit)[-1].rsplit('.',
                        maxsplit))
        if name not in seen or maxsplit >= splits:
            seen.add(name)
            return name
        return generate_name(path, maxsplit + 1, splits)

    if settings['PLUGINS']:
        return [generate_name(path) for path in settings['PLUGINS']]
    else:
        return ['Annotations']


def include_exclude_filter(include_patterns, exclude_patterns):
    filterf = None
    includef = None
    if include_patterns:
        pattern = include_patterns[0] if len(include_patterns) == 1 else \
            "(?:%s)" % '|'.join(include_patterns)
        includef = re.compile(pattern).search
        filterf = includef
    if exclude_patterns:
        pattern = exclude_patterns[0] if len(exclude_patterns) == 1 else \
            "(?:%s)" % '|'.join(exclude_patterns)
        excludef = re.compile(pattern).search
        if not includef:
            filterf = lambda x: not excludef(x)
        else:
            filterf = lambda x: includef(x) and not excludef(x)
    return filterf if filterf else bool


class IndexedDict(OrderedDict):
    """
    Ordered dictionary where values can also be obtained by their index as if
    they were in a list

    >>> idd = IndexedDict([('spam', 1), ('eggs', 2), ('bacon', 3)])
    >>> idd['spam']
    1
    >>> idd[0]
    1
    >>> idd['bacon']
    3
    >>> idd[2]
    3
    >>> idd[2] = 'ham'
    Traceback (most recent call last):
        ...
    TypeError: keys must not be an integers
    >>> idd[3]
    Traceback (most recent call last):
        ...
    IndexError: index out of range
    """
    def __setitem__(self, key, value):
        if isinstance(key, int):
            raise TypeError("keys must not be an integers")
        super(IndexedDict, self).__setitem__(key, value)

    def __getitem__(self, key):
        if isinstance(key, int):
            if key >= len(self):
                raise IndexError('index out of range')
            for i, k in enumerate(self):
                if i == key:
                    key = k
                    break
        return super(IndexedDict, self).__getitem__(key)


def _quotify(mystr):
    """
    quotifies an html tag attribute value.
    Assumes then, that any ocurrence of ' or " in the
    string is escaped if original string was quoted
    with it.
    So this function does not altere the original string
    except for quotation at both ends, and is limited just
    to guess if string must be quoted with '"' or "'"
    """
    quote = '"'
    l = len(mystr)
    for i in range(l):
        if mystr[i] == "\\" and i + 1 < l and mystr[i + 1] == "'":
            quote = "'"
            break
        elif mystr[i] == "\\" and i + 1 < l and mystr[i + 1] == '"':
            quote = '"'
            break
        elif mystr[i] == "'":
            quote = '"'
            break
        elif mystr[i] == '"':
            quote = "'"
            break
    return quote + mystr + quote


def serialize_tag(tag):
    """
    Converts a tag into a string when a slice [tag.start:tag.end]
    over the source can't be used because tag has been modified
    """
    out = "<"
    if tag.tag_type == HtmlTagType.CLOSE_TAG:
        out += "/"
    out += tag.tag

    attributes = []
    for key, val in tag.attributes.items():
        aout = key
        if val is not None:
            aout += "=" + _quotify(val)
        attributes.append(aout)
    if attributes:
        out += " " + " ".join(attributes)

    if tag.tag_type == HtmlTagType.UNPAIRED_TAG:
        out += "/"
    return out + ">"


def _must_add_tagid(element):
    return (hasattr(element, 'tag_type') and
            hasattr(element, 'tag') and
            element.tag_type != CLOSE_TAG and
            element.tag != 'ins')


def _modify_tagids(source, add=True):
    """Add or remove tags ids to/from HTML document"""
    output = []
    tagcount = 0
    if not isinstance(source, HtmlPage):
        source = HtmlPage(body=source)
    for element in source.parsed_body:
        if _must_add_tagid(element):
            if add:
                element.attributes[TAGID] = str(tagcount)
                tagcount += 1
            else:  # Remove previously added tagid
                element.attributes.pop(TAGID, None)
            output.append(serialize_tag(element))
        else:
            output.append(source.body[element.start:element.end])
    return u''.join(output)


def add_tagids(source):
    """
    Applies a unique attribute code number for each tag element in order to be
    identified later in the process of apply annotation"""
    return _modify_tagids(source, True)


def remove_tagids(source):
    """remove from the given page, all tagids previously added by add_tagids()
    """
    return _modify_tagids(source, False)


class Storage(object):
    def __init__(self, base_path):
        self.base_path = os.path.abspath(base_path)

    def rel_path(self, *args):
        return os.sep.join(args)

    def _path(self, *args):
        return os.path.join(self.base_path, self.rel_path(*args))

    def isdir(self, *args, **kwargs):
        return os.path.isdir(self._path(*args), **kwargs)

    def listdir(self, *args, **kwargs):
        return os.listdir(self._path(*args), **kwargs)

    def open(self, *args, **kwargs):
        """Open files from filesystem."""
        raw = kwargs.pop('raw', False)
        with open(self._path(*args)) as f:
            return decode(f.read()) if raw else json.load(f)


class SpiderLoader(object):
    def __init__(self, storage):
        if isinstance(storage, six.string_types):
            self.storage = Storage(storage)
        else:
            fsattrs = ['isdir', 'listdir', 'open', 'rel_path']
            if any(not hasattr(storage, attr) for attr in fsattrs):
                raise TypeError('Storage class must have "{}" methods'.format(
                    '", "'.join(fsattrs)))
            self.storage = storage
        self.spider_dir = self.storage.rel_path('spiders')
        self.spider_names = {
            s[:-len('.json')] for s in self.storage.listdir(self.spider_dir)
            if s.endswith('.json')
        }
        self._spiders = {}

    def __getitem__(self, key):
        if key not in self.spider_names:
            raise KeyError('The spider "{}" does not exist'.format(key))
        if key not in self._spiders:
            self._spiders[key] = self.load_spider(key)
        return self._spiders[key]

    def load_spider(self, spider_name):
        spec = self.storage.open(self.spider_dir,
                                 '{}.json'.format(spider_name))
        try:
            if spec.get('templates'):
                templates = []
                for template in spec.get('templates', []):
                    if template.get('version', '') < '0.13.0':
                        templates.append(template)
                    else:
                        templates.append(_build_sample(template))
                spec['templates'] = templates
            else:
                templates = self.load_external_templates(self.spider_dir,
                                                         spider_name)
                spec.setdefault("templates", []).extend(templates)
            return spec
        except ValueError as e:
            raise ValueError(
                "Error parsing spider (invalid JSON): %s: %s" %
                (spider_name, e)
            )

    def keys(self):
        for spider_name in self.spider_names:
            yield spider_name

    def items(self):
        spiders = chain(self._spiders, self.spider_names - set(self._spiders))
        for spider_name in spiders:
            yield spider_name, self[spider_name]

    def values(self):
        for _, spider in self.items():
            yield spider

    def load_external_templates(self, spec_base, spider_name):
        """A generator yielding the content of all passed `template_names` for
        `spider_name`.
        """
        spider_dir = self.storage.rel_path('spiders', spider_name)
        if not self.storage.isdir(spider_dir):
            raise StopIteration
        for name in self.storage.listdir(spider_dir):
            if not name.endswith('.json'):
                continue
            path = self.storage.rel_path(spider_dir, name)
            sample = self.storage.open(path)
            if not sample:
                continue
            sample_dir = path[:-len('.json')]
            if self.storage.isdir(sample_dir):
                for fname in self.storage.listdir(sample_dir):
                    if fname.endswith('.html'):
                        attr = fname[:-len('.html')]
                        html = self.storage.open(sample_dir, fname, raw=1)
                        sample[attr] = html
            if 'original_body' not in sample:
                sample['original_body'] = u'<html></html>'
            version = sample.get('version', '')
            yield _build_sample(sample, legacy=version < '0.13.0')



import unittest

from slybot.plugins.scrapely_annotations.migration import handle_tables


class MigrationTests(unittest.TestCase):
    def test_table_generalization(self):
        selectors = [
            ('table.mainbg > tr:nth-child(5) > td:nth-child(1) > '
             'table:nth-child(2) > tr:nth-child(3) > td:nth-child(1) > '
             'p:nth-child(1) > strong:nth-child(1)',
             'table.mainbg > tr:nth-child(5) > td:nth-child(1) > '
             'table:nth-child(2) > tr:nth-child(3) > td:nth-child(1) > '
             'p:nth-child(1) > strong:nth-child(1), table.mainbg > * > '
             'tr:nth-child(5) > td:nth-child(1) > table:nth-child(2) > '
             'tr:nth-child(3) > td:nth-child(1) > p:nth-child(1) > '
             'strong:nth-child(1), table.mainbg > tr:nth-child(5) > '
             'td:nth-child(1) > table:nth-child(2) > * > tr:nth-child(3) > '
             'td:nth-child(1) > p:nth-child(1) > strong:nth-child(1), '
             'table.mainbg > * > tr:nth-child(5) > td:nth-child(1) > '
             'table:nth-child(2) > * > tr:nth-child(3) > td:nth-child(1) > '
             'p:nth-child(1) > strong:nth-child(1)'),
            ('div > p > .table > div > span', 'div > p > .table > div > span'),
            ('div > p > #table > div > span', 'div > p > #table > div > span'),
            ('div > p > table > div > span',
             'div > p > table > div > span, div > p > table > * > div > span'),
            ('div > p > table:nth-child(4) > div > span',
             'div > p > table:nth-child(4) > div > span, '
             'div > p > table:nth-child(4) > * > div > span'),
            ('table', 'table')
        ]
        for selector, generalized in selectors:
            self.assertEqual(handle_tables(selector), generalized)



import json

import six

from scrapy.settings import Settings
from os.path import dirname

from scrapy.http.response.html import HtmlResponse
from scrapy.http.response.text import TextResponse
from scrapy.http.response.xml import XmlResponse
from scrapy.utils.spider import arg_to_iter

from scrapely.htmlpage import HtmlPage

from slybot.spider import IblSpider
from slybot.plugins.scrapely_annotations.builder import Annotations
from slybot.utils import read


PATH = dirname(__file__)


def request_to_set(requests):
    return [(set(t[0]),) + t[1:] for t in requests]


class UTF8Response(TextResponse):
    def __init__(self, *args, **kwargs):
        body = kwargs.get('body', args[3] if len(args) >= 4 else b'')
        if 'encoding' not in kwargs and isinstance(body, six.text_type):
            kwargs['encoding'] = 'utf-8'
        super(UTF8Response, self).__init__(*args, **kwargs)


class UTF8HtmlResponse(UTF8Response, HtmlResponse):
    pass


class UTF8TextResponse(UTF8Response, TextResponse):
    pass


class UTF8XmlResponse(UTF8Response, XmlResponse):
    pass


def make_spider(start_urls=None, sample=None):
    sample = [] if sample is None else arg_to_iter(sample)
    start_urls = [] if start_urls is None else arg_to_iter(start_urls)
    return {'start_urls': start_urls, 'templates': sample}


def open_spec(name):
    use_json = True if name.endswith('.json') else False
    with open('%s/data/templates/%s' % (PATH, name)) as f:
        return json.load(f) if use_json else read(f)


def open_sample_and_page(name):
    sample_spec = open_spec(name)
    url = sample_spec['url']
    return (HtmlPage(url=url, body=Annotations(sample_spec).apply()),
            HtmlPage(url=url, body=sample_spec['original_body']))


def open_page(name):
    return HtmlResponse(url=name, body=open_spec(name), encoding='utf-8')


def open_spider_page_and_results(name):
    sample_spec = open_spec(name)
    schemas = sample_spec['schemas']
    results = sample_spec['results']
    if 'original_body' not in sample_spec:
        sample_spec['original_body'] = open_spec(
            '{}.html'.format(name[:-len('.json')]))
    page = UTF8HtmlResponse('http://url', body=sample_spec['original_body'])
    spider = IblSpider(name, make_spider(sample=sample_spec), schemas, {},
                       Settings())
    return spider, page, results
